import pandas as pd
import time as time

from tqdm import tqdm   
from ..datamodel import Data, PYJEDAIFeature
from ..evaluation import Evaluation
from ..workflow import PYJEDAIWorkFlow, BlockingBasedWorkFlow
from ..clustering import AbstractClustering, UniqueMappingClustering, ConnectedComponentsClustering
from ..block_building import StandardBlocking
from ..block_cleaning import BlockFiltering, BlockPurging
from ..comparison_cleaning import WeightedEdgePruning, CardinalityEdgePruning, WeightedNodePruning
from ..matching import EntityMatching

from abc import abstractmethod

from typing import Optional, List, Tuple

class AbstractSchemaClustering(AbstractClustering):
    """Abstract class for schema clustering methods
    """
        
    def __init__(self):
        super().__init__()
        self.execution_time: float = 0.0
        self.schema_clustering_execution_time: float = 0.0

    @abstractmethod
    def _configuration(self) -> dict:
        pass

    @abstractmethod
    def stats(self) -> None:
        pass
    
    def report(self) -> None:
        """Prints method configuration
        """
        print(
            "Method name: " + self._method_name +
            "\nMethod info: " + self._method_info +
            ("\nParameters: \n" + ''.join(['\t{0}: {1}\n'.format(k, v) for k, v in self._configuration().items()]) if self._configuration().items() else "\nParameters: Parameter-Free method\n") +
            "\nRuntime (schema-clustering): {:2.4f} seconds".format(self.schema_clustering_execution_time) +
            "\nRuntime (total): {:2.4f} seconds".format(self.execution_time)
        )

class SchemaClustering(AbstractSchemaClustering):
    """Class to provide schema clustering methods
    """
    _method_name = "Schema Clustering"
    _method_info = "Performs pyjedai workflow to the names or values of the given data schema and then for each cluster performs pyjedai workflow for entity resolution"

    def __init__(self):
        super().__init__()
        self.on: str = 'names'
        self.schema_clustering_workflow: str
        self.entity_resolution_workflow: str

    def _configuration(self) -> dict:
        return {
            'on': self.on,
            'schema_clustering_workflow': self.schema_clustering_workflow,
            'entity_resolution_workflow': self.entity_resolution_workflow
        }

    def transform_mapping_to_ids(self,
                                 clusters,
                                 data: Data) -> pd.DataFrame:
        pairs_dict = dict()
        for cluster in clusters:
            lcluster = list(cluster)
            for i1 in range(0, len(lcluster)):
                for i2 in range(i1+1, len(lcluster)):
                    if lcluster[i1] < data.dataset_limit:
                        id1 = data._gt_to_ids_reversed_1[lcluster[i1]]
                        id2 = data._gt_to_ids_reversed_1[lcluster[i2]] \
                                if data.is_dirty_er \
                                else data._gt_to_ids_reversed_2[lcluster[i2]]
                    else:
                        id2 = data._gt_to_ids_reversed_2[lcluster[i1]]
                        id1 = data._gt_to_ids_reversed_1[lcluster[i2]]

                    pairs_dict[str(id1)] = str(id2)
        return pairs_dict

    def stats(self) -> None:
        print(
            "Statistics:" +
            "\n\tNew subsets shapes: " + \
            ''.join(['\n\t\tSubset {0}: {1}'.format(i, shape) for i, shape in self.new_data_stats]))
        return {
            self.new_data_stats    
        }

    def evaluate(self, 
                 prediction, 
                 export_to_df: bool = False, 
                 export_to_dict: bool = False, 
                 with_classification_report: bool = False, 
                 verbose: bool = True) -> any:

        if self.data is None:
            raise AttributeError("Can not proceed to evaluation without data object.")

        if self.data.ground_truth is None:
            raise AttributeError("Can not proceed to evaluation without a ground-truth file. " + 
                    "Data object has not been initialized with the ground-truth file")

        eval_obj = Evaluation(self.data)
        true_positives = 0
        
        for _, (id1, id2) in tqdm(self.data.ground_truth.iterrows(), desc="Evaluating"):
            if id1 in prediction and  \
                    prediction[id1] == id2:
                true_positives += 1
        eval_obj.calculate_scores(true_positives=true_positives, total_matching_pairs=len(self.data.ground_truth))
        return eval_obj.report(self.method_configuration(),
                                export_to_df,
                                export_to_dict,
                                with_classification_report,
                                verbose)

    def process(self, 
                data: Data, on='names', 
                pyjedai_workflow_for_clustering: PYJEDAIWorkFlow = None,
                pyjedai_workflow_for_er: PYJEDAIWorkFlow = None,
                verbose_schema_clustering: bool = False,
                verbose_er: bool = False,
                return_clusters = False) -> Optional[List[Tuple[pd.DataFrame, pd.DataFrame]]]:

        _start_time = time.time()
        self.data = data
        self.on = on

        if pyjedai_workflow_for_clustering == None:
            pyjedai_workflow_for_clustering = BlockingBasedWorkFlow()
            if data.is_dirty_er:
                pyjedai_workflow_for_clustering.default_schema_clustering_workflow_der()
            else:
                pyjedai_workflow_for_clustering.default_schema_clustering_workflow_ccer()

        self.schema_clustering_workflow = pyjedai_workflow_for_clustering.name
        
        print("Workflow used for schema clustering:", pyjedai_workflow_for_clustering.name)
        # D1
        entities_d1 = dict()
        for column in data.dataset_1.columns:
            if on == 'names':
                entities_d1[column] = column
            elif on == 'values':
                entities_d1[column] = ' '.join(data.dataset_1[column].astype(str))
            elif on == 'hybrid':
                entities_d1[column] = column + ' ' + ' '.join(data.dataset_1[column].astype(str))
            else:
                raise ValueError("on parameter must be one of 'names', 'values' or 'hybrid'")

        entities_d2 = None
        if not data.is_dirty_er:
            entities_d2 = dict()
            # D2
            for column in data.dataset_2.columns:
                if on == 'names':
                    entities_d2[column] = column
                elif on == 'values':
                    entities_d2[column] = ' '.join(data.dataset_2[column].astype(str))
                elif on == 'hybrid':
                    entities_d2[column] = column + ' ' + ' '.join(data.dataset_2[column].astype(str))
                else:
                    raise ValueError("on parameter must be one of 'names', 'values' or 'hybrid'")

        # Create dataframes from dictionaries
        attributes_d1 = pd.DataFrame.from_dict(entities_d1, orient='index', columns=['attribute'])
        attributes_d1['ids'] = range(0, len(attributes_d1))

        if not data.is_dirty_er:
            attributes_d2 = pd.DataFrame.from_dict(entities_d2, orient='index', columns=['attribute'])
            attributes_d2['ids'] = range(0, len(attributes_d2))

        # Clustering with pyJedAI        
        attributes_data = Data(
            dataset_1=attributes_d1,
            attributes_1=['attribute'],
            id_column_name_1='ids',
            dataset_2=attributes_d2 if not data.is_dirty_er else None,
            attributes_2=['attribute'] if not data.is_dirty_er else None,
            id_column_name_2='ids' if not data.is_dirty_er else None
        )
        
        pyjedai_workflow_for_clustering.run(attributes_data, verbose=verbose_schema_clustering)

        clusters = pyjedai_workflow_for_clustering.clusters

        # Create a new cluster of entities not in clusters
        def find_entities_not_in_clusters(ids, clusters):
            all_entities = set(ids)
            entities_in_clusters = set.union(*clusters)
            entities_not_in_clusters = all_entities - entities_in_clusters
            new_set = set(entities_not_in_clusters)
            return new_set

        all_ids = set(range(0, len(attributes_data.dataset_1) + len(attributes_data.dataset_2)))
        redundant_entities = find_entities_not_in_clusters(all_ids, clusters)
        if len(redundant_entities) > 0:
            clusters.append(redundant_entities)

        def contains_attributes_from_both(limit, cluster):
            has_entity_from_d1 = any(num < limit for num in cluster)
            has_entity_from_d2 = any(num >= limit for num in cluster)

            return has_entity_from_d1 and has_entity_from_d2
        
        # print("Clusters: ", clusters)
        
        new_datasets = []
        for i, cluster in enumerate(clusters):

            if not contains_attributes_from_both(attributes_data.dataset_limit, cluster):
                continue
            
            non_nan_indexes_d1 = set()
            non_nan_indexes_d2 = set() if not data.is_dirty_er else None
            
            for entity in cluster:
                if entity < attributes_data.dataset_limit:
                    attribute_name = attributes_data.dataset_1.iloc[entity].name
                    new_ids_d1 = set(data.dataset_1[data.dataset_1[attribute_name].notna()].index)
                    non_nan_indexes_d1.update(new_ids_d1)
                else:
                    attribute_name = attributes_data.dataset_2.iloc[entity-attributes_data.dataset_limit].name
                    non_nan_indexes_d2_attr = set(data.dataset_2[data.dataset_2[attribute_name].notna()].index)
                    new_ids_2 = set(map(lambda x: x - attributes_data.dataset_limit, non_nan_indexes_d2_attr))
                    non_nan_indexes_d2.update(new_ids_2)
            
            new_df_1 = data.dataset_1.iloc[list(non_nan_indexes_d1)]
            
            if not data.is_dirty_er:
                new_df_2 = data.dataset_2.iloc[list(non_nan_indexes_d2)]
                new_datasets.append((new_df_1, new_df_2))
        
        self.schema_clustering_execution_time = time.time() - _start_time

        if return_clusters:
            return new_datasets
        
        if pyjedai_workflow_for_er == None:
            pyjedai_workflow_for_er = BlockingBasedWorkFlow()
            if data.is_dirty_er:
                pyjedai_workflow_for_er.best_blocking_workflow_der() 
            else:
                pyjedai_workflow_for_er.best_blocking_workflow_ccer()

        self.entity_resolution_workflow = pyjedai_workflow_for_er.name

        global_pairs_dict = dict()
        self.new_data_stats = []
        for i in tqdm(range(len(new_datasets)), desc="Entity resolution for clusters"):
            d1, d2 = new_datasets[i]
            self.new_data_stats.append((i, d1.shape, d2.shape))
            
            new_data = Data(
                dataset_1=d1,
                attributes_1=data.attributes_1,
                id_column_name_1=data.id_column_name_1,
                dataset_2=d2,
                attributes_2=data.attributes_2,
                id_column_name_2=data.id_column_name_2,
                ground_truth=data.ground_truth
            )
            new_data.print_specs()
            pyjedai_workflow_for_er.run(new_data, verbose=verbose_er)
            new_clusters = pyjedai_workflow_for_er.clusters
            pairs = self.transform_mapping_to_ids(new_clusters, new_data)            
            global_pairs_dict.update(pairs)
            
        self.execution_time = time.time() - _start_time
        
        return global_pairs_dict

class RDFSchemaClustering(AbstractSchemaClustering):
    """Class to provide schema clustering methods
    """
    _method_name = "RDF Schema Clustering"
    _method_info = "Performs pyjedai dirty er workflow for the concatenation of objects per predicate. " + \
                        "For each rdf-triple containing the corresponding predicate, a new set is created. " + \
                            "For each set, a pyjedai dirty er workflow runs and finds the duplicate entities. " + \
                                "Links between entities, give us the corresponding same subjects between datasets. "

    def __init__(self):
        super().__init__()
        self.on: str = 'predicates'
        self.schema_clustering_workflow: str
        self.entity_resolution_workflow: str

    def _configuration(self) -> dict:
        return {
            'on': self.on,
            'schema_clustering_workflow': self.schema_clustering_workflow,
            'entity_resolution_workflow': self.entity_resolution_workflow
        }

    def stats(self) -> None:
        pass

    def transform_mapping_to_ids(self,
                                 clusters,
                                 data: Data) -> pd.DataFrame:
        pairs_dict = dict()
        for cluster in clusters:
            lcluster = list(cluster)
            for i1 in range(0, len(lcluster)):
                for i2 in range(i1+1, len(lcluster)):
                    if lcluster[i1] < data.dataset_limit:
                        id1 = data._gt_to_ids_reversed_1[lcluster[i1]]
                        id2 = data._gt_to_ids_reversed_1[lcluster[i2]] \
                                if data.is_dirty_er \
                                else data._gt_to_ids_reversed_2[lcluster[i2]]
                    else:
                        id2 = data._gt_to_ids_reversed_2[lcluster[i1]]
                        id1 = data._gt_to_ids_reversed_1[lcluster[i2]]

                    sid1 = data.dataset_1.iloc[int(id1)].name
                    sid2 = data.dataset_2.iloc[int(id2)].name
                    pairs_dict[sid1] = sid2
        return pairs_dict

    def evaluate(self, 
                 prediction, 
                 export_to_df: bool = False, 
                 export_to_dict: bool = False, 
                 with_classification_report: bool = False, 
                 verbose: bool = True) -> any:

        if self.data is None:
            raise AttributeError("Can not proceed to evaluation without data object.")

        if self.data.ground_truth is None:
            raise AttributeError("Can not proceed to evaluation without a ground-truth file. " + 
                    "Data object has not been initialized with the ground-truth file")

        eval_obj = Evaluation(self.data)
        true_positives = 0
        i1=0
        i2=0
        
        for _, (id1, id2) in tqdm(self.data.ground_truth.iterrows(), desc="Evaluating"):
            # print("ids: ", id1," and ", id2)
            if id1 in prediction and prediction[id1] == id2:
                true_positives += 1
            if id1 in prediction:
                i1+=1
            if id2 in set(prediction.values()):
                i2+=1
                # print(id1, ":", id2)
                # print(prediction[id1], ":", id2)
        print("Num of-1: ", i1)
        print("Num of-2: ", i2)
        print("True Positives: ", true_positives)
        print("Total Matching Pairs: ", len(self.data.ground_truth))
        eval_obj.calculate_scores(true_positives=true_positives, total_matching_pairs=len(self.data.ground_truth))
        return eval_obj.report(self.method_configuration(),
                                export_to_df,
                                export_to_dict,
                                with_classification_report,
                                verbose)
    
    def process(self, 
                data: Data, 
                on='predicates', 
                pyjedai_workflow_for_clustering: PYJEDAIWorkFlow = None,
                pyjedai_workflow_for_er: PYJEDAIWorkFlow = None,
                verbose_schema_clustering: bool = False,
                verbose_er: bool = False,
                return_clusters = False) -> Optional[List[Tuple[pd.DataFrame, pd.DataFrame]]]:

        _start_time = time.time()
        self.data = data
        self.on = on
        # self._progress_bar = tqdm(total=100, desc="RDF Schema Clustering")
        print("Starting RDF Schema Clustering")
        if pyjedai_workflow_for_clustering == None:
            pyjedai_workflow_for_clustering = self.pyjedai_workflow_for_er_on_predicates()

        self.schema_clustering_workflow = pyjedai_workflow_for_clustering.name
        
        print("Workflow used for schema clustering:", pyjedai_workflow_for_clustering.name)
        
        # Concatenate dataframe predicates and join object values
        predicates = dict()
        predicates_set_d1 = set()
        print("D1: Concatenating predicates and joining object values")
        for _, row in tqdm(data.dataset_1.iterrows()):
            if row['predicate'] not in predicates:
                predicates[row['predicate']] = str()
            predicates[row['predicate']] += ' ' + row['object']
            predicates_set_d1.add(row['predicate'])
        
        print("D2: Concatenating predicates and joining object values")
        predicates_set_d2 = None
        if not data.is_dirty_er:
            predicates_set_d2 = set()
            for _, row in data.dataset_2.iterrows():
                if row['predicate'] not in predicates:
                    predicates[row['predicate']] = str()
                predicates[row['predicate']] += ' ' + row['object']
                predicates_set_d2.add(row['predicate'])
        
        # Create dataframe from predicate dictionary
        # print("Predicates: ", predicates)
        print("Creating dataframe from predicate dictionary")
        predicates_df = pd.DataFrame.from_dict(predicates, orient='index', columns=['concatenation'])
        predicates_df['ids'] = range(0, len(predicates_df))

        print(predicates_df.head())

        # Clustering with pyJedAI
        print("Clustering with pyJedAI")
        all_predicates_data = Data(
            dataset_1=predicates_df,
            attributes_1=['concatenation'],
            id_column_name_1='ids'
        )

        all_predicates_data.print_specs()
        
        pyjedai_workflow_for_clustering.run(data=all_predicates_data, 
                                            verbose=verbose_schema_clustering,
                                            workflow_step_tqdm_disable=not verbose_schema_clustering,
                                            workflow_tqdm_enable=verbose_schema_clustering)

        clusters = pyjedai_workflow_for_clustering.clusters
        
        # print("Clusters: ", clusters)

        # Create a new cluster of entities not in clusters
        def find_entities_not_in_clusters(ids, clusters):
            all_entities = set(ids)
            entities_in_clusters = set.union(*clusters)
            entities_not_in_clusters = all_entities - entities_in_clusters
            new_set = set(entities_not_in_clusters)
            return new_set

        all_ids = set(range(0, len(all_predicates_data.dataset_1)))
        redundant_entities = find_entities_not_in_clusters(all_ids, clusters)
        if len(redundant_entities) > 0:
            clusters.append(redundant_entities)

        print("\n\n\n Clusters: ", clusters)

        def contains_items_from_both(cluster):
            cluster = list(cluster)
            has_entity_from_d1 = any((all_predicates_data.dataset_1.iloc[cluster[i]].name in predicates_set_d1) for i in range(len(cluster)))
            has_entity_from_d2 = any(all_predicates_data.dataset_1.iloc[cluster[i]].name in predicates_set_d2 for i in range(len(cluster)))
            
            return has_entity_from_d1 and has_entity_from_d2
        
        new_datasets = []
        for i, cluster in tqdm(enumerate(clusters), desc="Creating subsets"):

            if not contains_items_from_both(cluster):
                continue
            
            non_nan_indexes_d1 = set()
            non_nan_indexes_d2 = set() if not data.is_dirty_er else None
            
            for predicate_id in cluster:
                
                predicate = all_predicates_data.dataset_1.iloc[predicate_id].name
                # print(predicate_id, ":", predicate)
                if predicate in predicates_set_d1:
                    new_ids_d1 = set(data.dataset_1[data.dataset_1['predicate'] == predicate].index)
                    non_nan_indexes_d1.update(new_ids_d1)
                
                if not data.is_dirty_er and predicate in predicates_set_d2:
                    new_ids_d2 = set(data.dataset_2[data.dataset_2['predicate'] == predicate].index)
                    non_nan_indexes_d2.update(new_ids_d2)

            new_df_1 = data.dataset_1.iloc[list(non_nan_indexes_d1)].copy()
                        
            if not data.is_dirty_er:
                new_df_2 = data.dataset_2.iloc[list(non_nan_indexes_d2)].copy()
                new_datasets.append((new_df_1, new_df_2))
                # print("New datasets: ", new_df_1.shape[0], " x ", new_df_2.shape[0])
            else:
                new_datasets.append(new_df_1)
        self.schema_clustering_execution_time = time.time() - _start_time

        # free(all)
        print("New datasets: ", len(new_datasets))
        
        if return_clusters:
            return new_datasets

        print("Freeing memory")
        del all_predicates_data, predicates_df, clusters, all_ids, \
            redundant_entities, non_nan_indexes_d1, non_nan_indexes_d2

        if pyjedai_workflow_for_er == None:
            pyjedai_workflow_for_er = self.pyjedai_workflow_for_er_on_subjects()
            # if data.is_dirty_er:
            #     pyjedai_workflow_for_er.best_blocking_workflow_der()
            # else:
            #     pyjedai_workflow_for_er.best_blocking_workflow_ccer()

        self.entity_resolution_workflow = pyjedai_workflow_for_er.name
        print("Workflow used for entity resolution between subjects:", pyjedai_workflow_for_er.name)

        global_pairs_dict = dict()

        for i in tqdm(range(len(new_datasets)), desc="Entity resolution for clusters"):
            d1, d2 = new_datasets[i]
            
            print("Dataset 1: ", d1.shape)
            # print(d1.head())

            print("Dataset 2: ", d2.shape)
            # print(d2.head())
            
            if d1.shape[0] < 2 or (not data.is_dirty_er and d2.shape[0] < 2):
                continue

            subjects_d1 = dict()
            for _, row in tqdm(d1.iterrows(), desc="Creating subjects for dataset 1"):
                if row['subject'] not in subjects_d1:
                    subjects_d1[row['subject']] = str()
                subjects_d1[row['subject']] += ' ' + row['object']
            
            subjects_d2 = None
            if not data.is_dirty_er:
                subjects_d2 = dict()
                for _, row in tqdm(d2.iterrows(), desc="Creating subjects for dataset 2"):
                    if row['subject'] not in subjects_d2:
                        subjects_d2[row['subject']] = str()
                    subjects_d2[row['subject']] += ' ' + row['object']
            
            subjects_df_1 = pd.DataFrame.from_dict(subjects_d1, orient='index', columns=['concatenation'])
            subjects_df_1['ids'] = range(0, len(subjects_df_1))
            print("Subjects df 1: ", subjects_df_1.shape)
            print(subjects_df_1)

            subjects_df_2 = pd.DataFrame.from_dict(subjects_d2, orient='index', columns=['concatenation'])
            subjects_df_2['ids'] = range(0, len(subjects_df_2))
            print("Subjects df 2: ", subjects_df_2.shape)
            print(subjects_df_2)



            # Clustering with pyJedAI
            ccer_in_subjects_data = Data(
                dataset_1=subjects_df_1,
                attributes_1=['concatenation'],
                id_column_name_1='ids',
                dataset_2=subjects_df_2,
                attributes_2=['concatenation'],
                id_column_name_2='ids'
            )
            
            print("Dataset 1: ", subjects_df_1.shape)
            # print(d1.head())

            print("Dataset 2: ", subjects_df_2.shape)
            # print(d2.head())
            
            # if subjects_df_1.shape[0] < 2 or (not data.is_dirty_er and subjects_df_1.shape[0] < 2):
            #     print("Skipping cluster ", i)
            #     continue

            print("Running pyJedAI workflow for entity resolution on subjects")
            ccer_in_subjects_data.print_specs()


            
            try:
                pyjedai_workflow_for_er.run(data=ccer_in_subjects_data, 
                                            verbose=verbose_er,
                                            workflow_step_tqdm_disable=not verbose_er,
                                            workflow_tqdm_enable=verbose_er)
                new_clusters = pyjedai_workflow_for_er.clusters
            except Exception as e:
                print("Error in cluster ", i)
                print("Error: ", e)
                continue
            
            # print("New: ", new_clusters)
            # print("Transforming mapping to ids")
            pairs = self.transform_mapping_to_ids(new_clusters, ccer_in_subjects_data)
            global_pairs_dict.update(pairs)
            # global_pairs_df = pd.concat([global_pairs_df, pairs], ignore_index=True)
            # all_clusters += new_clusters
            
            print("Finished cluster: ", i)
            # Print pairs
            print("Predicted pairs: ", len(global_pairs_dict))
            # loop
            for pair in global_pairs_dict:
                print(pair, ":", global_pairs_dict[pair])
            print("\n\n\n")
            # print("Global pairs: ", global_pairs_dict)
        
        # print(all_clusters)
        # print(global_pairs_dict)
        self.execution_time = time.time() - _start_time
        return global_pairs_dict
    
    def pyjedai_workflow_for_er_on_subjects(self) -> PYJEDAIWorkFlow:
        return BlockingBasedWorkFlow(
            block_building = dict(
                method=StandardBlocking,
            ),
            block_cleaning = [
                dict(
                    method=BlockFiltering,
                    params=dict(ratio=0.2)
                )
            ],
            comparison_cleaning = dict(method=WeightedNodePruning),
            entity_matching = dict(method=EntityMatching, params=dict(metric='cosine',
                                                                      tokenizer='char_tokenizer', 
                                                                      vectorizer='tfidf', 
                                                                      qgram=3, 
                                                                      similarity_threshold=0.0)),
            clustering = dict(method=UniqueMappingClustering, 
                              exec_params=dict(similarity_threshold=0.1)),
            name="subjects-ccer"
        )
    
    def pyjedai_workflow_for_er_on_predicates(self) -> PYJEDAIWorkFlow:
        return BlockingBasedWorkFlow(
            block_building = dict(
                method=StandardBlocking,
            ),
            block_cleaning = [
                dict(method=BlockPurging, params=dict(smoothing_factor=1.0)),
                dict(method=BlockFiltering)
            ],
            comparison_cleaning = dict(method=WeightedNodePruning),
            entity_matching = dict(method=EntityMatching,
                                   params=dict(metric='cosine',
                                               similarity_threshold=0.0)),
            clustering = dict(method=ConnectedComponentsClustering),
            name="predicates-er"
        )



import ollama
from .datamodel import Data, PYJEDAIFeature
from .evaluation import Evaluation
from abc import ABC, abstractmethod
import pandas as pd
from typing import Literal, Union
from ollama._types import ResponseError 
import networkx
from time import time
from tqdm import tqdm

from pyjedai.vector_based_blocking import EmbeddingsNNBlockBuilding

DEFAULT_SYSTEM_PROMPT = """You are given two record descriptions and your task is to identify
if the records refer to the same entity or not.

You must answer with just one word:
True. if the records are referring to the same entity,
False. if the records are referring to a different entity."""

class AbstrackLLMMatching(PYJEDAIFeature):
    """Abstract class for Matching with LLMs
    """
    
    def __init__(self) -> None:
        super().__init__()
    
    # @abstractmethod
    def _apply_main_processing(self) -> dict:
        pass

    @abstractmethod
    def _configuration(self) -> dict:
        pass
    
    def stats(self) -> None:
        pass
    
    def report(self) -> None:
        """Prints LLM Matching method configuration
        """
        print(
            "Method name: " + self._method_name +
            "\nMethod info: " + self._method_info +
            ("\nParameters: \n" + ''.join(['\t{0}: {1}\n'.format(k, v) for k, v in self._configuration().items()]) if self._configuration().items() else "\nParameters: Parameter-Free method\n") +
            "Attributes from D1:\n\t" + ', '.join(c for c in (self.attributes_1 if self.attributes_1 is not None \
                else self.data.dataset_1.columns)) +
            ("\nAttributes from D2:\n\t" + ', '.join(c for c in (self.attributes_2 if self.attributes_2 is not None \
                else self.data.dataset_2.columns)) if not self.data.is_dirty_er else "") +
            "\nRuntime: {:2.4f} seconds".format(self.execution_time)
        )


    def evaluate(self,
                 prediction,
                 export_to_df: bool = False,
                 export_to_dict: bool = False,
                 with_classification_report: bool = False,
                 verbose: bool = True) -> any:

        if self.data is None:
            raise AttributeError("Can not proceed to evaluation without data object.")

        if self.data.ground_truth is None:
            raise AttributeError("Can not proceed to evaluation without a ground-truth file. " +
                    "Data object has not been initialized with the ground-truth file")

        eval_obj = Evaluation(self.data)
        true_positives = 0
        total_matching_pairs = len(prediction)
        
        for _, (id1, id2) in self.data.ground_truth.iterrows():
            id1 = self.data._ids_mapping_1[id1]
            id2 = self.data._ids_mapping_1[id2] if self.data.is_dirty_er \
                                                else self.data._ids_mapping_2[id2]
            
            if (id1, id2) in prediction or   \
                (id2, id1) in prediction:
                true_positives += 1

        eval_obj.calculate_scores(true_positives=true_positives, 
                                  total_matching_pairs=total_matching_pairs)
        return eval_obj.report(self.method_configuration(),
                                export_to_df,
                                export_to_dict,
                                with_classification_report,
                                verbose)


    def export_to_df(self, prediction) -> pd.DataFrame:
        """creates a dataframe with the predicted pairs

        Args:
            prediction (any): Predicted candidate pairs

        Returns:
            pd.DataFrame: Dataframe with the predicted pairs
        """
        
        if self.data.is_dirty_er:
            pairs = [(self.data._gt_to_ids_reversed_1[id1], self.data._gt_to_ids_reversed_1[id2]) for id1, id2 in prediction]
        else:
            pairs = [(self.data._gt_to_ids_reversed_1[id1], self.data._gt_to_ids_reversed_2[id2]) for id1, id2 in prediction]
        
        pairs_df = pd.DataFrame(data=pairs, columns=['id1', 'id2'])
        
        return pairs_df
    
    
class OllamaMatching(AbstrackLLMMatching):
    """Ollama Matching
    """
    
    _method_name = "Ollama Matching"
    _method_short_name: str = "OLLM"
    _method_info = "Utilizes an LLM and predicts which pairs are matching"


    def __init__(self, llm_model: str):
        
        super().__init__()
        self.llm = llm_model
        self.execution_time = 0
        
        
        self.model_name : str
        self.candidate_pairs : list = list()

        try:
            if not llm_model in ollama.list()['models']:
                print(f"Pulling model {llm_model} from ollama")
                ollama.pull(llm_model)
        except ResponseError as e:
            print(e.__str__())
        

    def process(self,
            prediction: Union[networkx.Graph, dict],
            data: Data, 
            system_prompt: str = DEFAULT_SYSTEM_PROMPT,
            create_examples: bool = False,
            examples: dict = None,
            suffix: Literal ['z', 'ft','tf'] = 'z',
            attributes: any = None,
            tqdm_disable: bool = False) -> dict:
        
        start_time = time()
        self.data = data
        self.tqdm_disable, self.prediction, self.attributes = tqdm_disable, prediction, attributes
        self.system_prompt = f"{system_prompt}\n"
        
        if isinstance(prediction, networkx.Graph):
            self._extract_candidate_pairs_from_graph(prediction)
        else:
            self._extract_candidate_pairs_from_blocks(prediction)

        if suffix != 'z':  
            example_records = self._extract_example_records(examples, create_examples)
            example_cnt = 1
            
            if suffix == 'ft':
                if 'false' in example_records: 
                    for example in example_records['false']:
                            example_str = f"\nExample {example_cnt}\nrecord 1: {example[0]}\nrecord 2: {example[1]}\nAnswer: False."
                            example_cnt += 1
                            self.system_prompt = f"{self.system_prompt}{example_str}"
                if 'true' in example_records: 
                    for example in example_records['true']:
                            example_str = f"\nExample {example_cnt}\nrecord 1: {example[0]}\nrecord 2: {example[1]}\nAnswer: True."
                            example_cnt += 1
                            self.system_prompt = f"{self.system_prompt}{example_str}"
            else:
                if 'true' in example_records: 
                    for example in example_records['true']:
                            example_str = f"\nExample {example_cnt}\nrecord 1: {example[0]}\nrecord 2: {example[1]}\nAnswer: True."
                            example_cnt += 1
                            self.system_prompt = f"{self.system_prompt}{example_str}"                    
                if 'false' in example_records: 
                    for example in example_records['false']:
                            example_str = f"\nExample {example_cnt}\nrecord 1: {example[0]}\nrecord 2: {example[1]}\nAnswer: False."
                            example_cnt += 1
                            self.system_prompt = f"{self.system_prompt}{example_str}"
        
        self.model_name = f'{self.llm}-{suffix}'
        ollama.create(model=self.model_name, from_=self.llm, system=self.system_prompt)
        print(f"Created ollama model {self.model_name}")


        self._progress_bar_2 = tqdm(total = len(self.candidate_pairs),
                                  desc = f"{self._method_name} [{self.model_name}]",
                                  disable = self.tqdm_disable)


        self.pairs = list()
        for id1, id2 in self.candidate_pairs:
            r1, r2 = self._extract_records(id1, id2)
            

            query = f"record 1: {r1}, record 2: {r2}. Answer with True. or False."

            # return
            resp = ollama.chat(
                model = self.model_name,
                messages = [{'role': 'user', 'content': query}],
                options = {'stop': ['\n','.']},
                stream = False
            )
            
            self._progress_bar_2.update(1)
            
            if resp['message']['content'] == 'True':
                self.pairs.append((id1,id2))
            
            # responses.append(resp['message']['content'])

            # gt_value = 'True' if (dt1_id, dt2_id) in gt_set else 'False'
        
        self.execution_time = time() - start_time
        ollama.delete(self.model_name)


        
        
        
        return self.pairs
    
    def _extract_candidate_pairs_from_graph(self, prediction: networkx.Graph):
        uv_pairs = list(prediction.edges(data=False))
        for pair in uv_pairs: 
            index_1, index_2 = (pair[0], pair[1]) if pair[0] < len(self.data.dataset_1) else (pair[1], pair[0])
            self.candidate_pairs.append((index_1, index_2))
    
    
    def _extract_candidate_pairs_from_blocks(self, prediction: dict) -> None:
        self._progress_bar = tqdm(total = len(prediction),
                                  desc = f"{self._method_name} [Extracting Candidate Pairs]",
                                  disable = self.tqdm_disable)
        
        all_blocks = list(prediction.values())
        
        if 'Block' in str(type(all_blocks[0])):
            #extract from raw blocks
            self._extract_from_raw_blocks(prediction)
        elif isinstance(all_blocks[0], set):
            self._extract_from_candidate_pairs(prediction)
        else:
            raise AttributeError("Wrong type of Blocks")
        

    def _extract_from_raw_blocks(self, blocks: dict):
        if self.data.is_dirty_er:
            for _, block in blocks.items():
                entities_array = list(block.entities_D1)
                for index_1 in range(0, len(entities_array), 1):
                    for index_2 in range(index_1+1, len(entities_array), 1):
                        self.candidate_pairs.append((entities_array[index_1], entities_array[index_2]))
                self._progress_bar.update(1)
        else:
            for _, block in blocks.items():
                for entity_id1 in block.entities_D1:
                    for entity_id2 in block.entities_D2:
                        self.candidate_pairs.append((entity_id1, entity_id2))
                self._progress_bar.update(1)

    def _extract_from_candidate_pairs(self, blocks: dict):
        for entity_id, candidates in blocks.items():
            for candidate_id in candidates:
                self.candidate_pairs.append((entity_id, candidate_id))
            self._progress_bar.update(1)

    def _extract_records(self, id1: int, id2: int) -> tuple:
        id_column_1 = self.data.id_column_name_1
        id_column_2 = self.data.id_column_name_2 if not self.data.is_dirty_er else self.data.id_column_name_1

        
        if isinstance(self.attributes, dict):
            r1 = self.data.entities.iloc[id1][self.attributes.items()].str.cat(sep=' ').lower().strip()
            r2 = self.data.entities.iloc[id2][self.attributes.items()].str.cat(sep=' ').lower().strip()
        elif isinstance(self.attributes, list):
            r1 = self.data.entities.iloc[id1][self.attributes].str.cat(sep=' ').lower().strip()
            r2 = self.data.entities.iloc[id2][self.attributes].str.cat(sep=' ').lower().strip()
        else: 
            r1 = self.data.entities.iloc[id1].drop(id_column_1).str.cat(sep=' ').lower().strip()
            r2 = self.data.entities.iloc[id2].drop(id_column_2).str.cat(sep=' ').lower().strip()

        return (r1, r2)
   
    def _extract_example_records(self, examples: dict, create_examples: bool) -> dict:

        examples_records = {}

        if examples:
            if not ('true' or 'false' in examples):
                raise Exception("Must at least contains key with ids for 'true' or 'false'. (e.g \{'true': [(10, 5)], ...\})")

            if 'true' in examples:
                examples_records['true'] = []
                for true_1, true_2 in examples['true']:
                    id1 = self.data._ids_mapping_1[str(true_1)]
                    id2 = self.data._ids_mapping_1[str(true_2)] if self.data.is_dirty_er \
                                                        else self.data._ids_mapping_2[str(true_2)]
                    examples_records['true'].append(self._extract_records(id1, id2))

            if 'false' in examples:
                examples_records['false'] = []
                for false_1, false_2 in examples['false']:
                    id1 = self.data._ids_mapping_1[str(false_1)]
                    id2 = self.data._ids_mapping_1[str(false_2)] if self.data.is_dirty_er \
                                                        else self.data._ids_mapping_2[str(false_2)]
                    examples_records['false'].append(self._extract_records(id1, id2))
        elif create_examples:
            emb = EmbeddingsNNBlockBuilding(vectorizer='sdistilroberta', similarity_search='faiss')
            _, g = emb.build_blocks(data=self.data, num_of_clusters=1, top_k=1,similarity_distance = 'cosine',with_entity_matching=True, load_embeddings_if_exist=True)
            edges_by_weight = sorted(g.edges(data=True), key=lambda x: x[2]['weight'])
            uv_pairs = [(u, v) for u, v, _ in edges_by_weight]

            examples_records['true'] = []
            examples_records['false'] = []
            
            
            if not isinstance(self.data.ground_truth, pd.DataFrame):
                true_pair = uv_pairs[-1] 
                false_pair = uv_pairs[0]
                
                true_1, true_2 = (true_pair[0], true_pair[1]) if true_pair[0] < len(self.data.dataset_1) else (true_pair[1], true_pair[0])
                false_1, false_2 = (false_pair[0], false_pair[1]) if false_pair[0] < len(self.data.dataset_1) else (false_pair[1], false_pair[0])

                examples_records['true'].append(self._extract_records(true_1, true_2))
                examples_records['false'].append(self._extract_records(false_1, false_2))
            else:
                gt_set = self.data.ground_truth.to_records(index=False).tolist()
                for pair in uv_pairs:
                    index_1, index_2 = (pair[0], pair[1]) if pair[0] < len(self.data.dataset_1) else (pair[1], pair[0])
                    id1 = self.data._gt_to_ids_reversed_1[index_1]
                    id2 = self.data._gt_to_ids_reversed_2[index_2]
                    if (index_1, index_2) in self.candidate_pairs and (id1, id2) in gt_set:
                        examples_records['true'].append(self._extract_records(index_1, index_2))
                        break
                
                for pair in reversed(uv_pairs):
                    index_1, index_2 = (pair[0], pair[1]) if pair[0] < len(self.data.dataset_1) else (pair[1], pair[0])
                    id1 = self.data._gt_to_ids_reversed_1[index_1]
                    id2 = self.data._gt_to_ids_reversed_2[index_2]
                    if (index_1, index_2) in self.candidate_pairs and not (id1, id2) in gt_set:
                        examples_records['false'].append(self._extract_records(index_1, index_2))
                        break
  
        else:
            raise Exception("If suffix in ['ft', 'tf'] must provide a dict with the ids of the examples or set create_examples to True")
        return examples_records
            
    def _configuration(self):
        return {
            "LLM": self.model_name,
            "Prompt" : self.system_prompt,
            
        }
        
        
            
            




from pyjedai.datamodel import SpatialData, PYJEDAIFeature
from queue import PriorityQueue
from functools import reduce
from shapely import relate
from tqdm.auto import tqdm

DIMS = {
    'F': frozenset('F'),
    'T': frozenset('012'),
    '*': frozenset('F012'),
    '0': frozenset('0'),
    '1': frozenset('1'),
    '2': frozenset('2'),
    }

def pattern(pattern_string):
    return Pattern(pattern_string)

class Pattern(object):
    def __init__(self, pattern_string):
        self.pattern = tuple(pattern_string.upper())
    def __str__(self):
        return ''.join(self.pattern)
    def __repr__(self):
        return "DE-9IM pattern: '%s'" % str(self)
    def matches(self, matrix_string):
        matrix = tuple(matrix_string.upper())
        def onematch(p, m):
            return m in DIMS[p]
        return bool(
            reduce(lambda x, y: x * onematch(*y), zip(self.pattern, matrix), 1)
            )

class AntiPattern(object):
    def __init__(self, anti_pattern_string):
        self.anti_pattern = tuple(anti_pattern_string.upper())
    def __str__(self):
        return '!' + ''.join(self.anti_pattern)
    def __repr__(self):
        return "DE-9IM anti-pattern: '%s'" % str(self)
    def matches(self, matrix_string):
        matrix = tuple(matrix_string.upper())
        def onematch(p, m):
            return m in DIMS[p]
        return not (
            reduce(lambda x, y: x * onematch(*y),
                   zip(self.anti_pattern, matrix),
                   1)
            )

class NOrPattern(object):
    def __init__(self, pattern_strings):
        self.patterns = [tuple(s.upper()) for s in pattern_strings]
    def __str__(self):
        return '||'.join([''.join(list(s)) for s in self.patterns])
    def __repr__(self):
        return "DE-9IM or-pattern: '%s'" % str(self)
    def matches(self, matrix_string):
        matrix = tuple(matrix_string.upper())
        def onematch(p, m):
            return m in DIMS[p]
        for pattern in self.patterns:
            val = bool(
                reduce(lambda x, y: x * onematch(*y), zip(pattern, matrix), 1))
            if val is True:
                break
        return val

# Familiar names for patterns or patterns grouped in logical expression
# ---------------------------------------------------------------------
contains = Pattern('T*****FF*')
crosses_lines = Pattern('0********')   # cross_lines is only valid for pairs of lines and/or multi-lines
crosses_1 = Pattern('T*T******')       # valid for mixed types (dim(a) < dim(b))
crosses_2 = Pattern('T*****T**')       # valid for mixed types (dim(a) > dim(b))
disjoint = Pattern('FF*FF****')
equal = Pattern('T*F**FFF*')
intersects = AntiPattern('FF*FF****')
overlaps1 = Pattern('T*T***T**')   # points and regions share an overlap pattern
overlaps2 = Pattern('1*T***T**')   # valid for lines only
touches = NOrPattern(['FT*******', 'F**T*****', 'F***T****'])
within = Pattern('T*F**F***')
covered_by = NOrPattern(['T*F**F***','*TF**F***','**FT*F***','**F*TF***'])
covers =  NOrPattern(['T*****FF*',	'*T****FF*',	'***T**FF*',	'****T*FF*'])




class RelatedGeometries :
        def __init__(self, qualifyingPairs) :
            self.pgr = 0
            self.exceptions = 0
            self.detectedLinks = 0
            self.verifiedPairs = 0
            self.qualifyingPairs = qualifyingPairs
            self.interlinkedGeometries = 0
            self.continuous_unrelated_Pairs = 0
            self.violations = 0
            self.containsD1 = []
            self.containsD2 = []
            self.coveredByD1 = []
            self.coveredByD2 = []
            self.coversD1 = []
            self.coversD2 = []
            self.crossesD1 = []
            self.crossesD2 = []
            self.equalsD1 = []
            self.equalsD2 = []
            self.intersectsD1 = []
            self.intersectsD2 = []
            self.overlapsD1 = []
            self.overlapsD2 = []
            self.touchesD1 = []
            self.touchesD2 = []
            self.withinD1 = []
            self.withinD2 = []

        def addContains(self, gId1,  gId2) :
          self.containsD1.append(gId1)
          self.containsD2.append(gId2)
        def addCoveredBy(self, gId1,  gId2):
           self.coveredByD1.append(gId1)
           self.coveredByD2.append(gId2)
        def addCovers(self, gId1,  gId2):
           self.coversD1.append(gId1)
           self.coversD2.append(gId2)
        def addCrosses(self, gId1,  gId2) :
          self.crossesD1.append(gId1)
          self.crossesD2.append(gId2)
        def addEquals(self, gId1,  gId2) :
          self.equalsD1.append(gId1)
          self.equalsD2.append(gId2)
        def addIntersects(self, gId1,  gId2) :
          self.intersectsD1.append(gId1)
          self.intersectsD2.append(gId2)
        def addOverlaps(self, gId1,  gId2) :
          self.overlapsD1.append(gId1)
          self.overlapsD2.append(gId2)
        def addTouches(self, gId1,  gId2) :
          self.touchesD1.append(gId1)
          self.touchesD2.append(gId2)
        def addWithin(self, gId1,  gId2) :
          self.withinD1.append(gId1)
          self.withinD2.append(gId2)

        def  getInterlinkedPairs(self) :
            return self.interlinkedGeometries
        def  getNoOfContains(self) :
            return len(self.containsD1)
        def  getNoOfCoveredBy(self) :
            return len(self.coveredByD1)
        def  getNoOfCovers(self) :
            return len(self.coversD1)
        def  getNoOfCrosses(self) :
            return len(self.crossesD1)
        def  getNoOfEquals(self) :
            return len(self.equalsD1)
        def  getNoOfIntersects(self) :
            return len(self.intersectsD1)
        def  getNoOfOverlaps(self) :
            return len(self.overlapsD1)
        def  getNoOfTouches(self) :
            return len(self.touchesD1)
        def  getNoOfWithin(self) :
            return len(self.withinD1)
        def  getVerifiedPairs(self) :
            return self.verifiedPairs

        def print(self) :
            print("Qualifying pairs:\t", str(self.qualifyingPairs))
            print("Exceptions:\t", str(self.exceptions))
            print("Detected Links:\t", str(self.detectedLinks))
            print("Interlinked geometries:\t", str(self.interlinkedGeometries))
            print("No of contains:\t", str(self.getNoOfContains()))
            print("No of covered-by:\t" + str(self.getNoOfCoveredBy()))
            print("No of covers:\t", str(self.getNoOfCovers()))
            print("No of crosses:\t", str(self.getNoOfCrosses()))
            print("No of equals:\t", str(self.getNoOfEquals()))
            print("No of intersects:\t" + str(self.getNoOfIntersects()))
            print("No of overlaps:\t", str(self.getNoOfOverlaps()))
            print("No of touches:\t", str(self.getNoOfTouches()))
            print("No of within:\t", str(self.getNoOfWithin()))

        def  verifyRelations(self, geomId1,  geomId2,  sourceGeom,  targetGeom) :
            related = False
            array = relate(sourceGeom, targetGeom)
            self.verifiedPairs += 1

            if intersects.matches(array):
                related = True
                self.detectedLinks += 1
                self.addIntersects(geomId1, geomId2)
            if within.matches(array):
                related = True
                self.detectedLinks += 1
                self.addWithin(geomId1, geomId2)
            if covered_by.matches(array):
                related = True
                self.detectedLinks += 1
                self.addCoveredBy(geomId1, geomId2)
            if crosses_lines.matches(array) or crosses_1.matches(array) or crosses_2.matches(array):
                related = True
                self.detectedLinks += 1
                self.addCrosses(geomId1, geomId2)
            if overlaps1.matches(array) or overlaps2.matches(array):
                related = True
                self.detectedLinks += 1
                self.addOverlaps(geomId1, geomId2)
            if  equal.matches(array):
                related = True
                self.detectedLinks += 1
                self.addEquals(geomId1, geomId2)
            if  touches.matches(array):
                related = True
                self.detectedLinks += 1
                self.addTouches(geomId1, geomId2)
            if  contains.matches(array):
                related = True
                self.detectedLinks += 1
                self.addContains(geomId1, geomId2)
            if covers.matches(array):
                related = True
                self.detectedLinks += 1
                self.addCovers(geomId1, geomId2)

            if (related) :
                self.interlinkedGeometries += 1
                self.pgr += self.interlinkedGeometries
                self.continuous_unrelated_Pairs = 0
            else:
                self.continuous_unrelated_Pairs += 1


            return related





class AbstractSpatialVerification(PYJEDAIFeature):
    """Abstract class for the block building method
    """
    def __init__(self, q_pairs):
        super().__init__()
        self.relations = RelatedGeometries(q_pairs)
        return
    
    def evaluate(self,
                 verbose: bool = True,
                ) -> any:
        if self.relations.qualifyingPairs != 0:
            self.recall_score = self.relations.interlinkedGeometries / float(self.relations.qualifyingPairs)
            self.precision_score = self.relations.interlinkedGeometries / self.relations.verifiedPairs
            self.f1_score = (2 * self.precision_score * self.recall_score) / (self.precision_score + self.recall_score)
            self.progressive_geometry_recall = self.relations.pgr / self.relations.qualifyingPairs / self.relations.verifiedPairs

            if(verbose):
                print(u'\u2500' * 123)
                print("Performance:\n\tPrecision: {:9.2f}% \n\tRecall:    {:9.2f}%\n\tF1-score:  {:9.2f}%\n\tProgressive Geometry Recall: {:9.2f}%".format(self.precision_score, self.recall_score, self.f1_score, self.progressive_geometry_recall))
                print(u'\u2500' * 123)

        return{"Precision:":self.precision_score,
               "Recall:":self.recall_score,
               "F1-score:":self.f1_score,
               "Progressivr geometry recall:":self.progressive_geometry_recall
              }

    def stats(self) -> None:
        self.relations.print()
        return

class StandardSpatialVerification(AbstractSpatialVerification):
    def __init__(self, q_pairs:int):
        super().__init__(q_pairs)
        self.q_pairs = q_pairs
        return

    def process(self, spatial_data:SpatialData, priority_pairs:PriorityQueue, verbose=True):
        counter = 0
        while(not priority_pairs.empty()):
            counter += 1
            weight, source_id, target_id, tEntity = priority_pairs.get()
            self.relations.verifyRelations(source_id, target_id, spatial_data.source_geometries[source_id], tEntity)

        # self.evaluate(verbose=True)

    
    def evaluate(self, verbose: bool = True) -> any:
        return super().evaluate(verbose)

    def _configuration(self) -> dict:
        """No configuration"""
        return {
            "Q-pairs": self.q_pairs
        }


"""Entity Matching Module
"""
import statistics
import pandas as pd
from time import time

import matplotlib.pyplot as plt
import numpy as np
from networkx import Graph
from .string_matchers import WhitespaceTokenizer, Cosine, Jaccard, GeneralizedJaccard, Dice, OverlapCoefficient
from stringcompare import Levenshtein, Jaro
from stringcompare.preprocessing import NGramTokenizer
from tqdm.autonotebook import tqdm

from .datamodel import Data, PYJEDAIFeature
from .evaluation import Evaluation
from .utils import WordQgramTokenizer, cosine, get_qgram_from_tokenizer_name, FrequencyEvaluator


metrics_mapping = {
    'edit_distance': Levenshtein(),
    'cosine' : Cosine(),
    'jaro' : Jaro(),
    'jaccard' : Jaccard(),
    'generalized_jaccard' : GeneralizedJaccard(),
    'dice': Dice(),
    'overlap_coefficient' : OverlapCoefficient(),
}

vector_metrics_mapping = {
    'cosine': cosine
}

string_metrics = [
    'jaro', 'edit_distance'
]

set_metrics = [
    'cosine', 'dice', 'generalized_jaccard', 'jaccard', 'overlap_coefficient'
]

vector_metrics = [ 
    'cosine', 'dice', 'jaccard', 'sqeuclidean'
]

whoosh_index_metrics = [
    'TF-IDF', 'Frequency', 'PL2', 'BM25F'
] 

faiss_metrics = [
    'cosine', 'euclidean'
]

magellan_metrics = string_metrics + set_metrics
available_metrics = magellan_metrics + vector_metrics + whoosh_index_metrics + faiss_metrics

#
# Tokenizers
#
# char_qgram_tokenizers = { 'char_'+ str(i) + 'gram':i for i in range(1, 7) }
# word_qgram_tokenizers = { 'word_'+ str(i) + 'gram':i for i in range(1, 7) }
char_qgram_tokenizers = ['char_tokenizer']
word_qgram_tokenizers = ['word_tokenizer']
magellan_tokenizers = ['white_space_tokenizer']
joins_tokenizers = ["qgrams", "standard", "standard_multiset", "qgrams_multiset"]

# tfidf_tokenizers = [ 'tfidf_' + cq for cq in char_qgram_tokenizers.keys() ] + \
#                     [ 'tfidf_' + wq for wq in word_qgram_tokenizers.keys() ]

# tf_tokenizers = [ 'tf_' + cq for cq in char_qgram_tokenizers.keys() ] + \
#                     [ 'tf_' + wq for wq in word_qgram_tokenizers.keys() ]
                        
# boolean_tokenizers = [ 'boolean_' + cq for cq in char_qgram_tokenizers.keys() ] + \
#                         [ 'boolean_' + wq for wq in word_qgram_tokenizers.keys() ]

# vector_tokenizers = tfidf_tokenizers + tf_tokenizers + boolean_tokenizers

# available_tokenizers = [key for key in char_qgram_tokenizers] + [key for key in word_qgram_tokenizers] + magellan_tokenizers + vector_tokenizers
available_tokenizers = char_qgram_tokenizers + word_qgram_tokenizers + magellan_tokenizers + joins_tokenizers
available_vectorizers = ['tfidf', 'tf', 'boolean']

class AbstractEntityMatching(PYJEDAIFeature):
    """Calculates similarity from 0.0 to 1.0
    """

    _method_name: str = "Abstract Entity Matching"
    _method_info: str = "Calculates similarity from 0. to 1."

    def __init__(
            self,
            metric: str = 'dice',
            similarity_threshold: float = 0.5,
        ) -> None:
        self.pairs: Graph
        self.metric = metric
        self.similarity_threshold = similarity_threshold
        self.execution_time = 0
        self.qgram=None
        #
        # Selecting tokenizer
        #
        if metric not in available_metrics:
            raise AttributeError(
                'Metric ({}) does not exist. Please select one of the available. ({})'.format(
                    metric, available_metrics
                )
            )
        else:
            self._metric = metric

    def _predict_candidate_pairs(self, blocks: dict) -> None:
        """Similarity evaluation after comparison cleaning.

        Args:
            blocks (dict): Comparison cleaning blocks.
        """
        for entity_id, candidates in blocks.items():
            for candidate_id in candidates:
                similarity = self._similarity(entity_id, candidate_id)
                self._insert_to_graph(entity_id, candidate_id, similarity)
            self._progress_bar.update(1)

    def _insert_to_graph(self, entity_id1, entity_id2, similarity):
        if self.similarity_threshold is None or \
            (self.similarity_threshold is not None and similarity > self.similarity_threshold):
            self.pairs.add_edge(entity_id1, entity_id2, weight=similarity)

    def report(self) -> None:
        """Prints Block Building method configuration
        """
        print(
            "Method name: " + self._method_name +
            "\nMethod info: " + self._method_info +
            ("\nParameters: \n" + ''.join(['\t{0}: {1}\n'.format(k, v) for k, v in self._configuration().items()]) if self._configuration().items() else "\nParameters: Parameter-Free method\n") +
            "Attributes:\n\t" + ', '.join(c for c in (self.attributes if self.attributes is not None \
                else self.data.dataset_1.columns)) +
            "\nRuntime: {:2.4f} seconds".format(self.execution_time)
        )

    def get_weights_avg(self) -> float:
        return sum([w for _, _, w in self.pairs.edges(data='weight')])/len(self.pairs.edges(data='weight'))

    def get_weights_median(self) -> float:
        return [w for _, _, w in sorted(self.pairs.edges(data='weight'))][int(len(self.pairs.edges(data='weight'))/2)]    
    
    def get_weights_standard_deviation(self) -> float:
        return statistics.stdev([w for _, _, w in self.pairs.edges(data='weight')])
    
    def plot_distribution_of_all_weights(self, save_figure_path=None) -> None:
        title = "Distribution of scores with " + self.metric + " metric in graph from entity matching"
        plt.figure(figsize=(10, 6))
        all_weights = [w for _, _, w in self.pairs.edges(data='weight')]
        sorted_weights = sorted(all_weights, reverse=True)
        
        plt.hist(sorted_weights)
        plt.xlim(0, 1)
        # only one line may be specified; full height
        plt.axvline(x = self.get_weights_avg(), color = 'blue', label = 'Average weight')
        plt.axvline(x = self.get_weights_median(), color = 'black', label = 'Median weight')
        plt.axvline(x = self.get_weights_avg()+self.get_weights_standard_deviation(), color = 'green', label = 'Average + SD weight')
        plt.legend()
        if save_figure_path:
            plt.savefig(save_figure_path)
        plt.show()

    def plot_distribution_of_all_weights_2d(self, save_figure_path=None) -> None:
        title = "Distribution of scores with " + self.metric + " metric in graph from entity matching"
        plt.figure(figsize=(10, 6))
        all_weights = [w for _, _, w in self.pairs.edges(data='weight')]
        sorted_weights = sorted(all_weights, reverse=True)
        
        fig, ax = plt.subplots(tight_layout=True)
        hist = ax.hist2d(sorted_weights, sorted_weights)
        plt.axvline(x = self.get_weights_avg(), color = 'blue', label = 'Average weight')
        plt.axvline(x = self.get_weights_median(), color = 'black', label = 'Median weight')
        plt.axvline(x = self.get_weights_avg()+self.get_weights_standard_deviation(), color = 'green', label = 'Average + SD weight')
        plt.legend()
        if save_figure_path:
            plt.savefig(save_figure_path)
        plt.show()

    def plot_distribution_of_scores(self, save_figure_path=None) -> None:
        title = "Distribution of scores with " + self.metric + " metric in graph from entity matching"
        def weight_distribution(G):
            bins = [0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0]
            distribution = [0] * (len(bins)-1)
            for u, v, w in G.edges(data='weight'):
                for i in range(len(bins) - 1):
                    if bins[i] <= w < bins[i + 1]:
                        distribution[i] += 1
                        break
            return distribution, len(G.edges(data='weight'))

        labels = [f'{(i)/10:.1f} - {(i+1)/10:.1f}' for i in range(0, 10)]

        distribution, num_of_pairs = weight_distribution(self.pairs)
        width = 0.5
        x = np.arange(len(labels))  # the label locations
        distribution = list(map(lambda x: (x/num_of_pairs)*100, distribution))
        print("Distribution-% of predicted scores: ", distribution)

        fig, ax = plt.subplots(figsize=(10,6))
        r1 = ax.bar(x, distribution, width, align='center', color='red')
        ax.set_xticks(x)
        ax.set_xticklabels(labels)
    
        
        # Add some text for labels, title and custom x-axis tick labels, etc.
        ax.set_ylabel('Percentage of pairs in each range to all (%)')
        ax.set_title(title)
        ax.set_xlabel('Similarity score range')
        fig.tight_layout()
        
        # only one line may be specified; full height
        plt.axvline(x = self.get_weights_avg()*10, color = 'blue', label = 'Average weight')
        plt.axvline(x = self.get_weights_median()*10, color = 'black', label = 'Median weight')
        plt.axvline(x = self.get_weights_avg()*10+self.get_weights_standard_deviation()*10, color = 'green', label = 'Average + SD weight')
        plt.legend()
        if save_figure_path:
            plt.savefig(save_figure_path)
        plt.show()

    def plot_gt_distribution_of_scores(self, save_figure_path=None) -> None:
        title = "Distribution of scores with " + self.metric + " metric on ground truth pairs"
        def weight_distribution():
            bins = [0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0]
            distribution = [0] * (len(bins)-1)
            for _, (id1, id2) in self.data.ground_truth.iterrows():
                id1 = self.data._ids_mapping_1[id1]
                id2 = self.data._ids_mapping_1[id2] if self.data.is_dirty_er else self.data._ids_mapping_2[id2]
                w = self._calculate_vector_similarity(id1, id2)

                for i in range(len(bins) - 1):
                    if bins[i] <= w < bins[i + 1]:
                        distribution[i] += 1
                        break
            return distribution, len(self.data.ground_truth)

        labels = [f'{(i)/10:.1f} - {(i+1)/10:.1f}' for i in range(0, 10)]

        distribution, num_of_pairs = weight_distribution()
        width = 0.5
        x = np.arange(len(labels))  # the label locations
        distribution = list(map(lambda x: (x/num_of_pairs)*100, distribution))
        print("Distribution-% of predicted scores: ", distribution)

        fig, ax = plt.subplots(figsize=(10,6))
        r1 = ax.bar(x, distribution, width, align='center', color='blue')
        ax.set_xticks(x)
        ax.set_xticklabels(labels)

        # Add some text for labels, title and custom x-axis tick labels, etc.
        ax.set_ylabel('Percentage of pairs in each range to all (%)')
        ax.set_title(title)
        ax.set_xlabel('Similarity score range')
        fig.tight_layout()
        if save_figure_path:
            plt.savefig(save_figure_path)
        plt.show()

    def evaluate(self,
                 prediction,
                 export_to_df: bool = False,
                 export_to_dict: bool = False,
                 with_classification_report: bool = False,
                 verbose: bool = True) -> any:

        if self.data is None:
            raise AttributeError("Can not proceed to evaluation without data object.")

        if self.data.ground_truth is None:
            raise AttributeError("Can not proceed to evaluation without a ground-truth file. " +
                    "Data object has not been initialized with the ground-truth file")

        eval_obj = Evaluation(self.data)
        true_positives = 0
        total_matching_pairs = prediction.number_of_edges()
        for _, (id1, id2) in self.data.ground_truth.iterrows():
            id1 = self.data._ids_mapping_1[id1]
            id2 = self.data._ids_mapping_1[id2] if self.data.is_dirty_er \
                                                else self.data._ids_mapping_2[id2]
            if (id1 in prediction and id2 in prediction[id1]) or   \
                 (id2 in prediction and id1 in prediction[id2]):
                true_positives += 1

        eval_obj.calculate_scores(true_positives=true_positives,
                                  total_matching_pairs=total_matching_pairs)
        return eval_obj.report(self.method_configuration(),
                                export_to_df,
                                export_to_dict,
                                with_classification_report,
                                verbose)
        
    def stats(self) -> None:
        pass
        
    def export_to_df(self, prediction: Graph, tqdm_enable=False) -> pd.DataFrame:
        """Creates a dataframe with the predicted pairs.

        Args:
            prediction (Graph): Predicted graph
            tqdm_enable (bool): Whether to enable tqdm progress bar

        Returns:
            pd.DataFrame: Dataframe with the predicted pairs
        """
        pairs_list = []

        is_dirty_er = self.data.is_dirty_er
        dataset_limit = self.data.dataset_limit
        gt_to_ids_reversed_1 = self.data._gt_to_ids_reversed_1
        if not is_dirty_er:
            gt_to_ids_reversed_2 = self.data._gt_to_ids_reversed_2

        for edge in tqdm(prediction.edges, disable=not tqdm_enable, desc="Exporting to DataFrame"):
            node1, node2 = edge

            if not is_dirty_er:
                if node1 < dataset_limit:
                    id1 = gt_to_ids_reversed_1[node1]
                    id2 = gt_to_ids_reversed_2[node2]
                else:
                    id1 = gt_to_ids_reversed_1[node2]
                    id2 = gt_to_ids_reversed_2[node1]
            else:
                id1 = gt_to_ids_reversed_1[node1]
                id2 = gt_to_ids_reversed_1[node2]

            pairs_list.append((id1, id2))

        pairs_df = pd.DataFrame(pairs_list, columns=['id1', 'id2'])

        return pairs_df

class EntityMatching(AbstractEntityMatching):
    """Calculates similarity from 0.0 to 1.0 for all blocks
    """

    _method_name: str = "Entity Matching"
    _method_info: str = "Calculates similarity from 0. to 1. for all blocks"

    def __init__(
            self,
            metric: str = 'dice',
            tokenizer: str = 'white_space_tokenizer',
            vectorizer : str = None,
            qgram : int = 1,
            similarity_threshold: float = 0.0,
            tokenizer_return_unique_values = False, # unique values or not,
            attributes: any = None,
        ) -> None:
        super().__init__()
        self.pairs: Graph
        self.metric = metric
        self.attributes: list = attributes
        self.similarity_threshold = similarity_threshold
        self.tokenizer = tokenizer
        self.execution_time = 0
        self.vectorizer = vectorizer
        self.qgram: int = -1
        #
        # Selecting tokenizer
        #
        if metric not in available_metrics:
            raise AttributeError(
                'Metric ({}) does not exist. Please select one of the available. ({})'.format(
                    metric, available_metrics
                )
            )
        else:
            self._metric = metric

        self.tokenizer_return_set = (metric in set_metrics) or tokenizer_return_unique_values    
        self.qgram : int = qgram
        
        if(vectorizer is not None):
            if self.vectorizer not in available_vectorizers:
                raise AttributeError(
                    'Weighting Scheme ({}) does not exist. Please select one of the available. ({})'.format(
                        vectorizer, available_vectorizers
                    )
                )
                
        elif(tokenizer is not None):
            if tokenizer == 'white_space_tokenizer':
                self._tokenizer = WhitespaceTokenizer()
            elif tokenizer == 'char_tokenizer':
                self._tokenizer = NGramTokenizer(n=self.qgram)
            elif tokenizer == 'word_tokenizer':
                self._tokenizer = WordQgramTokenizer(q=self.qgram)
            elif tokenizer not in available_tokenizers:
                raise AttributeError(
                    'Tokenizer ({}) does not exist. Please select one of the available. ({})'.format(
                        tokenizer, available_tokenizers
                    )
                )
        
    def predict(self,
                blocks: dict,
                data: Data,
                tqdm_disable: bool = False) -> Graph:
        """Main method of entity matching. Inputs a set of blocks and outputs a graph \
            that contains of the entity ids (nodes) and the similarity scores between them (edges).

            Args:
                blocks (dict): blocks of entities
                data (Data): dataset module
                tqdm_disable (bool, optional): Disables progress bar. Defaults to False.

            Returns:
                networkx.Graph: entity ids (nodes) and similarity scores between them (edges)
        """
        start_time = time()
        self.tqdm_disable = tqdm_disable
        
        if not blocks:
            raise ValueError("Empty blocks structure")
        self.data = data
        self.pairs = Graph()
        all_blocks = list(blocks.values())
        self._progress_bar = tqdm(total=len(blocks),
                                  desc=self._method_name+" ("+self.metric+ ", " + str(self.tokenizer) + ")",
                                  disable=self.tqdm_disable)

        if self.vectorizer is not None:
            self.initialize_vectorizer()

        if 'Block' in str(type(all_blocks[0])):
            self._predict_raw_blocks(blocks)
        elif isinstance(all_blocks[0], set):
            self._predict_candidate_pairs(blocks)
        else:
            raise AttributeError("Wrong type of Blocks")
        self.execution_time = time() - start_time
        self._progress_bar.close()

        return self.pairs
            
    def _predict_raw_blocks(self, blocks: dict) -> None:
        """Method for similarity evaluation blocks after Block building

        Args:
            blocks (dict): Block building blocks
        """
        if self.data.is_dirty_er:
            for _, block in blocks.items():
                entities_array = list(block.entities_D1)
                for index_1 in range(0, len(entities_array), 1):
                    for index_2 in range(index_1+1, len(entities_array), 1):
                        similarity = self._similarity(entities_array[index_1],
                                                      entities_array[index_2])
                        self._insert_to_graph(entities_array[index_1],
                                              entities_array[index_2],
                                              similarity)
                self._progress_bar.update(1)
        else:
            for _, block in blocks.items():
                for entity_id1 in block.entities_D1:
                    for entity_id2 in block.entities_D2:
                        similarity = self._similarity(entity_id1, entity_id2)
                        self._insert_to_graph(entity_id1, entity_id2, similarity)
                self._progress_bar.update(1)

    def initialize_vectorizer(self) -> None:
        self.frequency_evaluator : FrequencyEvaluator = FrequencyEvaluator(vectorizer=self.vectorizer,
                                                                            tokenizer=self.tokenizer,
                                                                            qgram=self.qgram)
        d1 = self.data.dataset_1[self.attributes] if self.attributes else self.data.dataset_1
        self._entities_d1 = d1 \
                    .apply(" ".join, axis=1) \
                    .apply(lambda x: x.lower()) \
                    .values.tolist()
        
        d2 = None
        if(not self.data.is_dirty_er):
            d2 = self.data.dataset_2
            if self.attributes:
                d2 = d2[self.attributes]

        self._entities_d2 = d2 \
                    .apply(" ".join, axis=1) \
                    .apply(lambda x: x.lower()) \
                    .values.tolist() if not self.data.is_dirty_er else self._entities_d1 
        
        
        _dataset_identifier : str = ('_'.join([self.data.dataset_name_1, self.data.dataset_name_2])) if(self.data.dataset_name_1 is not None and self.data.dataset_name_2 is not None) else ("dataset") 
        self.frequency_evaluator.fit(metric=self.metric,
                                    dataset_identifier=_dataset_identifier,
                                    indexing='inorder',
                                    d1_entities=self._entities_d1,
                                    d2_entities=self._entities_d2)

    def _similarity(self, entity_id1: int, entity_id2: int) -> float:

        similarity: float = 0.0
        if self.vectorizer is not None:
            return self.frequency_evaluator.predict(id1=entity_id1, id2=entity_id2)
        elif isinstance(self.attributes, dict):
            for attribute, weight in self.attributes.items():
                e1 = self.data.entities.iloc[entity_id1][attribute].lower()
                e2 = self.data.entities.iloc[entity_id2][attribute].lower()
                if self.tokenizer_return_set: 
                    similarity += weight*metrics_mapping[self._metric].compare(
                        set(self._tokenizer.tokenize(e1)) if self._metric in set_metrics else e1,
                        set(self._tokenizer.tokenize(e2)) if self._metric in set_metrics else e2
                    )
                else:     
                    similarity += weight*metrics_mapping[self._metric].compare(
                        self._tokenizer.tokenize(e1) if self._metric in set_metrics else e1,
                        self._tokenizer.tokenize(e2) if self._metric in set_metrics else e2
                    )
        elif isinstance(self.attributes, list):
            for attribute in self.attributes:
                e1 = self.data.entities.iloc[entity_id1][attribute].lower()
                e2 = self.data.entities.iloc[entity_id2][attribute].lower()
                if self.tokenizer_return_set: 
                    similarity += metrics_mapping[self._metric].compare(
                        set(self._tokenizer.tokenize(e1)) if self._metric in set_metrics else e1,
                        set(self._tokenizer.tokenize(e2)) if self._metric in set_metrics else e2
                    )
                else:
                    similarity += metrics_mapping[self._metric].compare(
                        self._tokenizer.tokenize(e1) if self._metric in set_metrics else e1,
                        self._tokenizer.tokenize(e2) if self._metric in set_metrics else e2
                    )
            similarity /= len(self.attributes)
        else:
            e1 = self.data.entities.iloc[entity_id1].str.cat(sep=' ').lower()
            e2 = self.data.entities.iloc[entity_id2].str.cat(sep=' ').lower()
            if self.tokenizer_return_set: 
                te1 = set(self._tokenizer.tokenize(e1)) if self._metric in set_metrics else e1
                te2 = set(self._tokenizer.tokenize(e2)) if self._metric in set_metrics else e2
            else:
                te1 = self._tokenizer.tokenize(e1) if self._metric in set_metrics else e1
                te2 = self._tokenizer.tokenize(e2) if self._metric in set_metrics else e2
            similarity = metrics_mapping[self._metric].compare(te1, te2)
        return similarity

    def _configuration(self) -> dict:
        return  {
            "Metric" : self.metric,
            "Attributes" : self.attributes,
            "Similarity threshold" : self.similarity_threshold,
            "Tokenizer" : self.tokenizer,
            "Vectorizer" : self.vectorizer if self.vectorizer is not None else "None",
            "Qgrams" : self.qgram
        }

class VectorBasedMatching(AbstractEntityMatching):

    _method_name: str = "Vector Based Matching"
    _method_info: str = "Calculates similarity from 0. to 1. for vectors"

    def __init__(
            self,
            metric: str = 'cosine',
            similarity_threshold: float = 0.5,
        ) -> None:
        self.pairs: Graph
        self.metric = metric
        self.similarity_threshold = similarity_threshold
        self.vectors_d1 = None
        self.vectors_d2 = None
        self.execution_time = 0

        #
        # Selecting tokenizer
        #
        if metric not in vector_metrics:
            raise AttributeError(
                'Metric ({}) does not exist. Please select one of the available. ({})'.format(
                    metric, available_metrics
                )
            )
        else:
            self._metric = metric
            
    def predict(self,
                blocks: dict,
                data: Data,
                vectors_d1: np.array,
                vectors_d2: np.array = None,
                tqdm_disable: bool = False,
        ) -> Graph:
        """Main method of entity matching. Inputs a set of blocks and outputs a graph \
            that contains of the entity ids (nodes) and the similarity scores between them (edges).

            Args:
                blocks (dict): blocks of entities
                data (Data): dataset module
                tqdm_disable (bool, optional): Disables progress bar. Defaults to False.

            Returns:
                networkx.Graph: entity ids (nodes) and similarity scores between them (edges)
        """
        start_time = time()
        self.tqdm_disable = tqdm_disable
        self.vectors_d1 = vectors_d1
        # self.vectors_d2 = vectors_d2
        

        if(vectors_d1 is None):
            raise ValueError("Embeddings of the first dataset not given")
        else:
            self.vectors = vectors_d1
            if(not data.is_dirty_er):
                if(vectors_d2 is None):
                    raise ValueError("Embeddings of the second dataset not given")
                self.vectors = np.concatenate((vectors_d1,vectors_d2), axis=0)
        self.data = data
        self.pairs = Graph()
        self._progress_bar = tqdm(total=len(blocks),
                                  desc=self._method_name+" ("+self.metric+ ", " + str(self.tokenizer) + ")",
                                  disable=self.tqdm_disable)
        self._predict_candidate_pairs(blocks)
        self.execution_time = time() - start_time
        self._progress_bar.close()

        return self.pairs

    def _similarity(self, entity_id1: int, entity_id2: int) -> float:
        return vector_metrics_mapping[self._metric](self.vectors[entity_id1], self.vectors[entity_id2])

    def _configuration(self) -> dict:
        conf =  {
            "Metric" : self.metric,
            "Similarity threshold" : self.similarity_threshold
        }
        
        return conf



"""
This code was based on py_stringmatching: https://github.com/anhaidgroup/py_stringmatching
"""
from abc import ABC, abstractmethod
from stringcompare import Jaro
import re


class WhitespaceTokenizer(ABC):
    def tokenize(self, sentence):
        whitespace_pattern = re.compile(r'\s+')
        tokens = whitespace_pattern.split(sentence.strip())
        tokens = [token for token in tokens if token]
        return tokens
    
class StringMatcher(ABC):
    """String Matchers based on py_stringmatching"""
    flag = True
    
    
    def check_instance_type(self, te1, te2) -> None:
        if not isinstance(te1, list) and not isinstance(te1, set): 
            raise TypeError("Must be either list or set")
        if not isinstance(te2, list) and not isinstance(te2, set): 
            raise TypeError("Must be either list or set")
        
    def exact_match(self, te1, te2):
        return te1 == te2
    
    def empty_match(self, te1, te2):
        return len(te1) == 0 or len(te2) == 0
            
        
    @abstractmethod
    def compare(self, te1, te2):
        pass


class Cosine(StringMatcher):
    def compare(self, te1, te2) -> float:

        self.check_instance_type(te1, te2)

        # if exact match return 1.0
        if self.exact_match(te1, te2):
            return 1.0

        # if one of the strings is empty return 0
        if self.empty_match(te1, te2):
            return 0.0

        intersection = len(set(te1) & set(te2))
        norm1 = len(te1) ** 0.5
        norm2 = len(te2) ** 0.5
        return intersection / (norm1 * norm2) if norm1 * norm2 > 0 else 0
        
class Dice(StringMatcher):
    def compare(self, te1, te2) -> float:
        self.check_instance_type(te1, te2)

        set1 = set(te1)
        set2 = set(te2)

        # if exact match return 1.0
        if self.exact_match(set1, set2):
            return 1.0

        # if one of the strings is empty return 0
        if self.empty_match(set1, set2):
            return 0.0

        return 2.0 * float(len(set1 & set2)) / float(len(set1) + len(set2))

class Jaccard(StringMatcher):
    def compare(self, te1, te2) -> float:
        self.check_instance_type(te1, te2)

        set1 = set(te1)
        set2 = set(te2)

        # if exact match return 1.0
        if self.exact_match(set1, set2):
            return 1.0

        # if one of the strings is empty return 0
        if self.empty_match(set1, set2):
            return 0.0

        intersection = len(set(te1) & set(te2))
        return intersection/(len(set1) + len(set2) + intersection)

class GeneralizedJaccard(StringMatcher):
    def compare(self, te1, te2) -> float:
        self.check_instance_type(te1, te2)

        set1 = set(te1)
        set2 = set(te2)

        # if exact match return 1.0
        if self.exact_match(set1, set2):
            return 1.0

        # if one of the strings is empty return 0
        if self.empty_match(set1, set2):
            return 0.0

        set1_x = set()
        set2_y = set()
        match_score = 0.0
        match_count = 0
        list_matches = []
        threshold=0.5
        for element in set1:
            for item in set2:
                score = Jaro().compare(element, item)
                if score > 1 or score < 0:
                    raise ValueError('Similarity measure should' + \
                                    ' return value in the range [0,1]')
                if score > threshold:
                    list_matches.append((element, item, score))

        # position of first string, second string and sim score in tuple
        first_string_pos = 0
        second_string_pos = 1
        sim_score_pos = 2

        # sort the score of all the pairs
        list_matches.sort(key=lambda x: x[sim_score_pos], reverse=True)

        # select score in increasing order of their weightage, 
        # do not reselect the same element from either set.
        for element in list_matches:
            if (element[first_string_pos] not in set1_x and
                element[second_string_pos] not in set2_y):
                set1_x.add(element[first_string_pos])
                set2_y.add(element[second_string_pos])
                match_score += element[sim_score_pos]
                match_count += 1

        return float(match_score) / float(len(set1) + len(set2) - match_count)

class OverlapCoefficient(StringMatcher):
    def compare(self, te1, te2) -> float:
        self.check_instance_type(te1, te2)

        set1 = set(te1)
        set2 = set(te2)

        # if exact match return 1.0
        if self.exact_match(set1, set2):
            return 1.0

        # if one of the strings is empty return 0
        if self.empty_match(set1, set2):
            return 0.0
        return float(len(set1 & set2)) / min(len(set1), len(set2))




'''
Contains all methods for creating embeddings from text 
and then performing NNs methods for cluster formation.
'''
import os
import pickle
import re
import sys
import warnings
import pandas as pd
from time import time
from typing import List, Literal, Tuple
from math import sqrt
import faiss
import platform
RUNNING_OS = platform.system()

try: 
    import gensim.downloader as api
    HAS_GENSIM = True
except ImportError:
    HAS_GENSIM = False
    

import networkx as nx
import numpy as np
import torch
import transformers
from sentence_transformers import SentenceTransformer
from tqdm.autonotebook import tqdm
from transformers import (
    AlbertModel,
    AlbertTokenizer,
    AutoModel,
    AutoTokenizer,
    BertModel,
    BertTokenizer,
    DistilBertModel,
    DistilBertTokenizer,
    RobertaModel,
    RobertaTokenizer,
    XLNetModel,
    XLNetTokenizer,
)
from math import log

transformers.logging.set_verbosity_error()
from faiss import normalize_L2

from .datamodel import Data, PYJEDAIFeature
from .evaluation import Evaluation
from .utils import SubsetIndexer

EMBEDDINGS_DIR = '.embs'
if not os.path.exists(EMBEDDINGS_DIR):
    os.makedirs(EMBEDDINGS_DIR)
    EMBEDDINGS_DIR = os.path.abspath(EMBEDDINGS_DIR)
    print('Created embeddings directory at: ' + EMBEDDINGS_DIR)


class EmbeddingsNNBlockBuilding(PYJEDAIFeature):
    """Block building via creation of embeddings and a Nearest Neighbor Approach.
    """

    _method_name = "Embeddings-NN Block Building"
    _method_info = "Creates a set of candidate pais for every entity id " + \
        "based on Embeddings creariot and Similarity search among the vectors."

    
    _gensim_mapping_download = {
        'fasttext' : 'fasttext-wiki-news-subwords-300',
        'glove' : 'glove-wiki-gigaword-300',
        'word2vec' : 'word2vec-google-news-300'
        
    } if HAS_GENSIM else {}
    
    _sentence_transformer_mapping = {
        'smpnet' : 'all-mpnet-base-v2',
        'st5' : 'gtr-t5-large',
        'sdistilroberta' : 'all-distilroberta-v1',
        'sminilm' : 'all-MiniLM-L12-v2',
        'sent_glove' : 'average_word_embeddings_glove.6B.300d'
    }

    def __init__(
            self,
            vectorizer: str,
            similarity_search: str = 'faiss'
    ) -> None:
        super().__init__()
        self.vectorizer, self.similarity_search = vectorizer, similarity_search
        self.embeddings: np.array
        self.vectors_1: np.array
        self.vectors_2: np.array = None
        self.vector_size: int
        self.num_of_clusters: int
        self.top_k: int
        self._faiss_metric_type = None

    def _tokenize_entity(self, entity: str) -> str:
        """Produces a list of workds of a given string

        Args:
            entity (str): String representation  of an entity

        Returns:
            str: entity string
        """
        return entity.strip().lower()#' '.join(list(filter(None, re.split('[\\W_]', entity.lower()))))

    def build_blocks(self,
                     data: Data,
                     vector_size: int = 300,
                     num_of_clusters: int = 5,
                     top_k: int = 30,
                     max_word_embeddings_size: int = 256,
                     attributes_1: list = None,
                     attributes_2: list = None,
                     return_vectors: bool = False,
                     tqdm_disable: bool = False,
                     save_embeddings: bool = True,
                     load_embeddings_if_exist: bool = False,
                     load_path_d1: str = None,
                     load_path_d2: str = None,
                     with_entity_matching: bool = False,
                     input_cleaned_blocks: dict = None,
                     similarity_distance: str = 'cosine',
                     custom_pretrained_model: Literal["word", "sentence"] = None,
                     verbose: bool = False
    ) -> any:
        """Main method of the vector based approach. Contains two steps. First an embedding method. \
            And afterwards a similarity search upon the vectors created in the previous step.
            Pre-trained schemes are used for the embedding process.

        Args:
            data (Data): dataset from datamodel
            vector_size (int, optional): For the Gensim vectorizers. Defaults to 300. \
                Qaution for the hugging face embeddings has no effect.
            num_of_clusters (int, optional): Number of clusters for FAISS. Defaults to 5.
            top_k (int, optional): Top K similar candidates. Defaults to 30.
            attributes_1 (list, optional): Vectorization of specific attributes for D1. \
                                            Defaults to None.
            attributes_2 (list, optional): Vectorization of specific attributes for D2. \
                                            Defaults to None.
            return_vectors (bool, optional): If true, returns the vectors created from the pretrained \
                                            embeddings instead of the blocks. Defaults to False.
            tqdm_disable (bool, optional): Disable progress bar. For experiment purposes. \
                                            Defaults to False.
            custom_pretrained_model (Literal["word", "sentence"], optional): If set, indicates \
                that a custom pre-trained model is used for word or sentence embeddings \

        Raises:
            AttributeError: Vectorizer check
            AttributeError: Similarity Search method check.

        Returns:
            dict: Entity ids to sets of top-K candidate ids. OR
            Tuple(np.array, np.array): vectors from d1 and vectors from d2
        """
        if self.similarity_search != 'faiss':
            raise AttributeError("Only FAISS is available for now.")
        
        _start_time = time()
        self.blocks = dict()
        self.verbose = verbose
        self.with_entity_matching = with_entity_matching
        self.save_embeddings, self.load_embeddings_if_exist = save_embeddings, load_embeddings_if_exist
        self.max_word_embeddings_size = max_word_embeddings_size
        self.similarity_distance = similarity_distance
        self.data, self.attributes_1, self.attributes_2, self.vector_size, self.num_of_clusters, self.top_k, self.input_cleaned_blocks \
            = data, attributes_1, attributes_2, vector_size, num_of_clusters, top_k, input_cleaned_blocks
        self.load_path_d1, self.load_path_d2 = load_path_d1, load_path_d2
        
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

        self._progress_bar = tqdm(total=data.num_of_entities,
                                  desc=(self._method_name + ' [' + self.vectorizer + ', ' + self.similarity_search + ', ' + str(self.device) + ']'),
                                  disable=tqdm_disable)
        self.custom_pretrained_model = custom_pretrained_model
        gensim_models = ['word2vec', 'fasttext', 'doc2vec', 'glove']
        word_emb_mdoels = ['bert', 'distilbert', 'roberta', 'xlnet', 'albert']
        sentence_emb_models = ['smpnet', 'st5', 'sent_glove', 'sdistilroberta', 'sminilm']
        reserved_model_names = gensim_models + word_emb_mdoels + sentence_emb_models

        if self.custom_pretrained_model and self.vectorizer in reserved_model_names:
            raise AttributeError(f"`custom_pretrained_model` is set, but the vectorizer is already reserved for a specific model: {self.vectorizer}. Please use a custom model, or set this to 'None'")
        
        if(input_cleaned_blocks == None):
            self._applied_to_subset = False
        else:
            _all_blocks = list(input_cleaned_blocks.values())
            if 'Block' in str(type(_all_blocks[0])):
                self._applied_to_subset = False
            elif isinstance(_all_blocks[0], set):
                self._applied_to_subset = True
            else:
                raise AttributeError("Wrong type of blocks given")

        self._si = SubsetIndexer(self.input_cleaned_blocks, self.data, self._applied_to_subset)
        self._d1_valid_indices: list[int] = self._si.d1_retained_ids
        self._d2_valid_indices: list[int] = [x - self.data.dataset_limit for x in self._si.d2_retained_ids]  if not data.is_dirty_er else None

        self._entities_d1 = data.dataset_1[attributes_1 if attributes_1 else data.attributes_1] \
                            .apply(" ".join, axis=1) \
                            .apply(self._tokenize_entity) \
                            .values.tolist()
        self._entities_d1 = [self._entities_d1[x] for x in self._d1_valid_indices]
        self._entities_d2 = data.dataset_2[attributes_2 if attributes_2 else data.attributes_2] \
                    .apply(" ".join, axis=1) \
                    .apply(self._tokenize_entity) \
                    .values.tolist() if not data.is_dirty_er else None
        self._entities_d2 = [self._entities_d2[x] for x in self._d2_valid_indices] if not data.is_dirty_er else None

        self.vectors_1 = None
        self.vectors_2 = None
        
        if self.with_entity_matching:
            self.graph = nx.Graph()
        
        self._d1_loaded : bool = False
        self._d2_loaded : bool = False
        if load_embeddings_if_exist:
            if verbose:
                print("Loading embeddings from file...")
            if(self.load_path_d1 is not None):
                p1 = self.load_path_d1
            else:
                p1 = os.path.join(EMBEDDINGS_DIR, self.vectorizer + '_' + (self.data.dataset_name_1 \
                                                    if self.data.dataset_name_1 is not None else "d1") +'.npy')
            if verbose:
                print("Attempting to load D1 embeddings...")
            if os.path.exists(p1):
                self.vectors_1 = vectors_1 = np.load(p1)
                self.vectors_1 = vectors_1 = vectors_1[self._d1_valid_indices]
                self._progress_bar.update(data.num_of_entities_1)
                self._d1_loaded = True
                if verbose:
                    print(f"{p1} -> Loaded Successfully")
            else:
                if verbose:
                    print("Embeddings not found for D1. Creating new ones.")
            
            if not data.is_dirty_er:
                if(self.load_path_d2 is not None):
                    p2 = self.load_path_d2
                else:
                    p2 = os.path.join(EMBEDDINGS_DIR, self.vectorizer + '_' + (self.data.dataset_name_2 \
                                                        if self.data.dataset_name_2 is not None else "d2") +'.npy')    
                if verbose: print("Attempting to load D2 embeddings...")
                if os.path.exists(p2):
                    self.vectors_2 = vectors_2 = np.load(p2)
                    self.vectors_2 = vectors_2 = vectors_2[self._d2_valid_indices]
                    self._progress_bar.update(data.num_of_entities_2)
                    self._d2_loaded = True
                    if verbose: print(f"{p2} -> Loaded Successfully")
                else:
                    if verbose: print("Embeddings not found for D2. Creating new ones.")

        if not self._d1_loaded or (not data.is_dirty_er and not self._d2_loaded):
            if self.vectorizer in ['word2vec', 'fasttext', 'doc2vec', 'glove']:
                self.vectors_1, self.vectors_2 = self._create_gensim_embeddings()
            elif self.vectorizer in ['bert', 'distilbert', 'roberta', 'xlnet', 'albert'] or (self.custom_pretrained_model == "word" and self.vectorizer not in reserved_model_names):
                self.vectors_1, self.vectors_2 = self._create_pretrained_word_embeddings()
            elif self.vectorizer in ['smpnet', 'st5', 'sent_glove', 'sdistilroberta', 'sminilm'] or (self.custom_pretrained_model == "sentence" and self.vectorizer not in reserved_model_names):
                self.vectors_1, self.vectors_2 = self._create_pretrained_sentence_embeddings()
            else:
                raise AttributeError("Not available vectorizer")
            
        if save_embeddings:
            if verbose: print("Saving embeddings...")
            
            if self._applied_to_subset:
                if verbose: print("Cannot save embeddings, subset embeddings storing not supported.")
            else:
                if not self._d1_loaded:
                    p1 = os.path.join(EMBEDDINGS_DIR, self.vectorizer + '_' + (self.data.dataset_name_1 \
                                                            if self.data.dataset_name_1 is not None else "d1") +'.npy')
                    if verbose: print("Saving file: ", p1)
                    np.save(p1, self.vectors_1)
                
                if not data.is_dirty_er and not self._d2_loaded:
                    p2 = os.path.join(EMBEDDINGS_DIR, self.vectorizer + '_' + (self.data.dataset_name_2 \
                                                            if self.data.dataset_name_2 is not None else "d2") +'.npy')
                    if verbose: print("Saving file: ", p2)
                    np.save(p2, self.vectors_2)

        if return_vectors:
            return (self.vectors_1, _) if data.is_dirty_er else (self.vectors_1, self.vectors_2)

        if self.similarity_search == 'faiss':
            if verbose: print("Starting similarity search with FAISS...")
            self._faiss_metric_type = faiss.METRIC_L2
            self._similarity_search_with_FAISS()
            if verbose: print("Similarity search with FAISS is done.")
        else:
            raise AttributeError("Not available method")
        self._progress_bar.close()        
        self.execution_time = time() - _start_time
        
        if self.with_entity_matching:
            return self.blocks, self.graph
        else:
            return self.blocks

    def _create_gensim_embeddings(self) -> Tuple[np.array, np.array]:
        """Embeddings with Gensim. More on https://github.com/RaRe-Technologies/gensim-data

        Args:
            entities_d1 (list): Entities from D1
            entities_d2 (list, optional): Entities from D2 (CCER). Defaults to None.

        Returns:
            Tuple[np.array, np.array]: Embeddings from D1 and D2
        """
        if not HAS_GENSIM : 
            raise ImportError("Reinstall pyjedai with pip install pyjedai[with-gensim] or use other vectorizer")
            
        vocabulary = api.load(self._gensim_mapping_download[self.vectorizer])
    
        if not self._d1_loaded:
            for e1 in self._entities_d1:
                vectors_1.append(self._create_vector(e1, vocabulary))
                self._progress_bar.update(1)
            vectors_1 = np.vstack(vectors_1).astype('float32')

        vectors_2 = []
        if not self.data.is_dirty_er and not self._d2_loaded:
            for e2 in self._entities_d2:
                vectors_2.append(self._create_vector(e2, vocabulary))
                self._progress_bar.update(1)
            vectors_2 = np.vstack(vectors_2).astype('float32')

        return vectors_1, vectors_2

    def _create_pretrained_word_embeddings(self) -> Tuple[np.array, np.array]:
        if self.vectorizer == 'bert':
            tokenizer = BertTokenizer.from_pretrained('bert-base-uncased')
            model = BertModel.from_pretrained("bert-base-uncased")
        elif self.vectorizer == 'distilbert':
            tokenizer = DistilBertTokenizer.from_pretrained('distilbert-base-uncased')
            model = DistilBertModel.from_pretrained("distilbert-base-uncased")
        elif self.vectorizer == 'roberta':
            tokenizer = RobertaTokenizer.from_pretrained('roberta-base')
            model = RobertaModel.from_pretrained('roberta-base')
        elif self.vectorizer == 'xlnet':
            tokenizer = XLNetTokenizer.from_pretrained('xlnet-base-cased')
            model = XLNetModel.from_pretrained('xlnet-base-cased')
        elif self.vectorizer == 'albert':
            tokenizer = AlbertTokenizer.from_pretrained('albert-base-v2')
            model = AlbertModel.from_pretrained("albert-base-v2")
        elif self.custom_pretrained_model == "word":
            tokenizer = AutoTokenizer.from_pretrained(self.vectorizer)
            model = AutoModel.from_pretrained(self.vectorizer)

        model = model.to(self.device)
        self.vectors_1 = self._transform_entities_to_word_embeddings(self._entities_d1,
                                                                     model,
                                                                     tokenizer) if not self._d1_loaded else self.vectors_1
        self.vector_size = self.vectors_1[0].shape[0]
        self.vectors_2 = self._transform_entities_to_word_embeddings(self._entities_d2,
                                                                     model,
                                                                     tokenizer) if not self.data.is_dirty_er and not self._d2_loaded else self.vectors_2
        return self.vectors_1, self.vectors_2

    def _transform_entities_to_word_embeddings(self, entities, model, tokenizer) -> np.array:
    
        model = model.to(self.device)
        embeddings = []
        
        for entity in entities:
            encoded_input = tokenizer(entity,
                                        return_tensors='pt',
                                        truncation=True,
                                        return_attention_mask = True,
                                        max_length=self.max_word_embeddings_size,
                                        padding='max_length')

            encoded_input = {key: value.to(self.device) for key, value in encoded_input.items()}  # Move input tensors to GPU

            with torch.no_grad():
                encoded_input = {key: value.to(self.device) for key, value in encoded_input.items()}  # Move input tensors to GPU
                output = model(**encoded_input)
                vector = output.last_hidden_state[:, 0, :]
                
            vector = vector.cpu().numpy()
            embeddings.append(vector.reshape(-1))
            self._progress_bar.update(1)
        
        self.vector_size = embeddings[0].shape[0]
        return np.array(embeddings).astype('float32')

    def _create_pretrained_sentence_embeddings(self):
        if self.vectorizer in self._sentence_transformer_mapping:
            model_name = self._sentence_transformer_mapping[self.vectorizer]
        elif self.custom_pretrained_model == "sentence":
            model_name = self.vectorizer
        else:
            raise AttributeError(f"unknown vectorizer: {self.vectorizer}")
        model = SentenceTransformer(model_name,
                                    device=self.device)
        vectors_1 = []
        if not self._d1_loaded:
            for e1 in self._entities_d1:
                vector = model.encode(e1)
                vectors_1.append(vector)
                self._progress_bar.update(1)
            self.vector_size = len(vectors_1[0])
            vectors_1 = np.vstack(vectors_1).astype('float32')
        vectors_2 = []
        if not self.data.is_dirty_er and not self._d2_loaded:            
            for e2 in self._entities_d2:
                vector = model.encode(e2)
                vectors_2.append(vector)
                self._progress_bar.update(1)
            self.vector_size = len(vectors_2[0])
            vectors_2 = np.vstack(vectors_2).astype('float32')
            
        return vectors_1, vectors_2 

    def _similarity_search_with_FAISS(self):
        if self.verbose:
            print("Creating index...")
        index = faiss.IndexFlatL2(self.vectors_1.shape[1])
        
        if self.similarity_distance == 'cosine' or self.similarity_distance == 'cosine_without_normalization':
            index.metric_type = faiss.METRIC_INNER_PRODUCT
        elif self.similarity_distance == 'euclidean':
            index.metric_type = faiss.METRIC_L2
        else:
            raise ValueError("Invalid similarity distance: ", self.similarity_distance)

        if self.similarity_distance == 'cosine':
            faiss.normalize_L2(self.vectors_1)
            if not self.data.is_dirty_er: faiss.normalize_L2(self.vectors_2)
        
        if self.verbose:
            print("Training index...")
        index.train(self.vectors_1)  # train on the vectors of dataset 1


        if self.similarity_distance == 'cosine':
            faiss.normalize_L2(self.vectors_1)
            if not self.data.is_dirty_er: faiss.normalize_L2(self.vectors_2)

        if self.verbose:
            print("Adding vectors to index...")
        index.add(self.vectors_1)   # add the vectors and update the index

        if self.similarity_distance == 'cosine':
            faiss.normalize_L2(self.vectors_1)
            if not self.data.is_dirty_er: faiss.normalize_L2(self.vectors_2)

        if self.verbose:
            print("Searching for neighbors...")
        self.distances, self.neighbors = index.search(self.vectors_1 if self.data.is_dirty_er else self.vectors_2, self.top_k)

        if self.similarity_distance == 'euclidean':
            self.distances = 1/(1 + self.distances)

        self.blocks = dict()
        if self.verbose:
            print("Building blocks...")

        for _entity in tqdm(range(0, self.neighbors.shape[0]), desc="Building blocks", disable=not self.verbose):
            
            _entity_id = self._si.d1_retained_ids[_entity] if self.data.is_dirty_er else self._si.d2_retained_ids[_entity]
            
            # if _entity_id not in self.blocks:
            #     self.blocks[_entity_id] = set()
            
            for _neighbor_index, _neighbor in enumerate(self.neighbors[_entity]):

                if _neighbor == -1:
                    continue
                
                _neighbor_id = self._si.d1_retained_ids[_neighbor]
                
                if _neighbor_id not in self.blocks:
                    self.blocks[_neighbor_id] = set()

                self.blocks[_neighbor_id].add(_entity_id)
                # self.blocks[_entity_id].add(_neighbor_id)
                
                if self.with_entity_matching:
                    self.graph.add_edge(_entity_id, _neighbor_id, weight=self.distances[_entity][_neighbor_index])

    def _create_vector(self, tokens: List[str], vocabulary) -> np.array:
        num_of_tokens = 0
        vector = np.zeros(self.vector_size)
        for token in tokens.split():
            if token in vocabulary:
                vector += vocabulary[token]
                num_of_tokens += 1
        if num_of_tokens > 0:
            vector /= num_of_tokens

        return vector

    def evaluate(self,
                 prediction,
                 export_to_df: bool = False,
                 export_to_dict: bool = False,
                 with_classification_report: bool = False,
                 verbose: bool = True,
                 with_stats: bool = False) -> any:

        if self.data is None:
            raise AttributeError("Can not proceed to evaluation without data object.")

        if self.data.ground_truth is None:
            raise AttributeError("Can not proceed to evaluation without a ground-truth file. " +
                    "Data object has not been initialized with the ground-truth file")

        eval_obj = Evaluation(self.data)
        true_positives = 0
        total_matching_pairs = sum([len(block) for block in prediction.values()])
        for _, (id1, id2) in self.data.ground_truth.iterrows():
            id1 = self.data._ids_mapping_1[id1]
            id2 = self.data._ids_mapping_1[id2] if self.data.is_dirty_er \
                                                else self.data._ids_mapping_2[id2]
            if (id1 in prediction and id2 in prediction[id1]) or   \
                (id2 in prediction and id1 in prediction[id2]):
                true_positives += 1

        eval_obj.calculate_scores(true_positives=true_positives, 
                                  total_matching_pairs=total_matching_pairs)
        evaluation = eval_obj.report(self.method_configuration(),
                                export_to_df,
                                export_to_dict,
                                with_classification_report,
                                verbose)
        
        if with_stats:
            self.stats()

        return evaluation

    def _configuration(self) -> dict:
        return {
            "Vectorizer" : self.vectorizer,
            "Similarity-Search" : self.similarity_search,
            "Top-K" : self.top_k,
            "Vector size": self.vector_size
        }
    
    def stats(self) -> None:
        print("Statistics:")
        if self.similarity_search == 'faiss':
            print(" FAISS:" +
                "\n\tIndices shape returned after search: " + str(self.neighbors.shape)
            )
        print(u'\u2500' * 123)

    def export_to_df(self, prediction: dict, tqdm_enable:bool = False) -> pd.DataFrame:
        """Creates a dataframe with the predicted pairs.

        Args:
            prediction (dict): Predicted candidate pairs.

        Returns:
            pd.DataFrame: Dataframe with the predicted pairs.
        """
        pairs_list = []

        is_dirty_er = self.data.is_dirty_er
        gt_to_ids_reversed_1 = self.data._gt_to_ids_reversed_1
        if not is_dirty_er:
            gt_to_ids_reversed_2 = self.data._gt_to_ids_reversed_2

        for entity_id, candidates in tqdm(prediction.items(), desc="Exporting to DataFrame", disable=not tqdm_enable):
            id1 = gt_to_ids_reversed_1[entity_id]

            for candidate_id in candidates:
                if is_dirty_er:
                    id2 = gt_to_ids_reversed_1[candidate_id]
                else:
                    id2 = gt_to_ids_reversed_2[candidate_id]

                pairs_list.append((id1, id2))

        pairs_df = pd.DataFrame(pairs_list, columns=['id1', 'id2'])

        return pairs_df
